/**
 * @file avplay.h
 * avplay public api header.
 */

#ifndef AVPLAY_H
#define AVPLAY_H


#ifdef __cplusplus
extern "C" {
#endif

#define AV_STRINGIFY(s)         AV_TOSTRING(s)
#define AV_TOSTRING(s) #s

#define LIBAVPLAY_VERSION_INT   ((49<<16)+(0<<8)+0)
#define LIBAVPLAY_VERSION       49.0.0
#define LIBAVPLAY_BUILD         LIBAVPLAY_VERSION_INT

#define LIBAVPLAY_IDENT         "Lavp" AV_STRINGIFY(LIBAVPLAY_VERSION)





#include <SDL.h>

/* typedefs for callback functions */

typedef void (*avplay_doneEvent_callback)(void *userData1, void *userData2);

typedef void (*avplay_video_refresh_callback)(int delay, void *is, void *userData);

typedef void (*avplay_audio_process_callback)(int16_t *audio_buf, int num_samples, int num_channels, int sample_rate,
	int do_reset, void *userData);

typedef struct {
  double masterClock; /* master playback clock in seconds */
  double clipStartTime; /* clip start time in seconds (0 if unknown) */
  double clipDuration;  /* clip duration in seconds (0 if unknown) */
  double av_diff;     /* audio clock - video clock */
  int aqsize;         /* audio queue size */
  int vqsize;         /* video queue size */
  int sqsize;         /* subtitle queue size */
} avplay_playback_status;
typedef void (*avplay_playback_status_callback)(avplay_playback_status *playback_status, void *userData);

typedef void (*avplay_handle_event_callback)(SDL_Event *e, void *userData);


/* must do this once before any other calls */
void avplay_init();


/* If your app runs an event loop in its main thread it must take 
   responsibility for handling video-related calls from that main thread.
   This simply requires implementing two callback functions, one for
   general video events and one for timed video refresh operations: */

/* The callback set here must wait for "delay" milliseconds, then
   cause a callback to avplay_video_refresh(is) from the main thread. 
   (Hint: use a timed event in your main thread event loop.) */
void avplay_set_video_refresh_callback(avplay_video_refresh_callback callback,
				       void *userData);
void avplay_video_refresh(void *is);
/* The callback set here must cause a callback to avplay_handle_event(e)
   from the main thread. (Hint: use your main thread event loop.) */
void avplay_set_handle_event_callback(avplay_handle_event_callback callback,
				      void *userData);
void avplay_handle_event(SDL_Event *e);


/* The callback set here can modify the audio samples in-place in the buffer */
void avplay_set_audio_process_callback(avplay_audio_process_callback callback,
				       void *userData);


/* This callback, if set, will get called with the status of the playback.
   (The callback must take care of calling avplay_free() with the 
   avplay_playback_status pointer that it is given.)
   You can use this for example to drive the state of a seekbar or
   do other actions based on the time. */
void avplay_set_playback_status_callback(avplay_playback_status_callback callback,
				       void *userData);

void avplay_free(void* p);

/* Call this to possibly resize/reposition the video rect to fit in the
   x/y/w/h provided.
   Returns nonzero if there was video available and the rect was 
   resized/repositioned based on the request and on the actual video
   aspect ratio.  The actual x/y/w/h used is returned in the pointer args. */
int avplay_resize_video_rect(int *pX, int *pY, int *pW, int *pH);

/* Play, stop, pause, seek. */

/* avplay_play() returns 0 if playback started OK, or -1 if you must
 *  call avplay_stop(), sleep for a short time (e.g. 10 ms), then try again.
 * For avplay_play(), if you want a callback made when playback is complete (i.e. end of file)
 *  then use doneEvent_callback to set a callback function with 2 arguments */
int avplay_play(const char* filename, void* SDL_windowId,
	avplay_doneEvent_callback doneEvent_callback, void *doneEvent_userData1, void *doneEvent_userData2);
void avplay_stop();
void avplay_toggle_pause();
void avplay_seek(double time);

#ifdef __cplusplus
}
#endif

#endif /* AVPLAY_H */



